#! /usr/bin/env perl
use Test2::V0;
use Language::FormulaEngine;
use Language::FormulaEngine::Parser;
use Language::FormulaEngine::Compiler;
use Language::FormulaEngine::Namespace;
use Language::FormulaEngine::Namespace::Default;
ok(1, 'all loaded');
done_testing;